#include "ds50daq/DAQ/V1495Driver.hh"

#include "artdaq/DAQdata/Debug.hh"
#include "artdaq/Application/GeneratorMacros.hh"
#include "cetlib/exception.h"
#include "darksidecore/Data/Config.hh"
#include "ds50daq/BuildInfo/GetPackageBuildInfo.hh"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "detail/caen_io.hh"

#include <algorithm>
#include <string>
#include <vector>
#include <cmath>

using fhicl::ParameterSet;
using ds50::V1495Fragment;

const std::string ds50::V1495Driver::FRAGMENTS_PROCESSED_STAT_KEY("V1495DriverFragmentsProcessed");
const std::string ds50::V1495Driver::READ_TIME_STAT_KEY("V1495DriverReadTime");
const std::string ds50::V1495Driver::FRAGMENT_CREATE_TIME_STAT_KEY("V1495DriverFragmentCreateTime");
const std::string ds50::V1495Driver::REGISTER_RW_TIME_STAT_KEY("V1495DriverRegisterReadWriteTime");
const std::string ds50::V1495Driver::IRQWAIT_TIME_STAT_KEY("V1495DriverIRQWaitTime");

const int ds50::V1495Driver::NO_CHOICE = -1;
const int ds50::V1495Driver::VETO_OFF = 0;
const int ds50::V1495Driver::ADC_SLAVE = 1;
const int ds50::V1495Driver::HYBRID = 2;

namespace V1495 {
  enum V1495_registers {
    fw_revision = 0x100C,
    port_a_mask = 0x1010,
    port_b_mask = 0x1014,
    run_control = 0x1018,
    window_inibit = 0x101C,
    mf_extention = 0x1020,
    trigger_control = 0x1024,
    run_number = 0x1028,
    test_pulser_dpm = 0x102C,
    status = 0x1030,
    test_pulser_read = 0x1034,
    tdc_delay = 0x1044,
    tdc_delay_veto = 0x102C,
    laser_control = 0x1048,
    fifo_depth = 0x104C,
    module_id = 0x1050,
    g2_gate = 0x1054,
    g2_threshold = 0x1058,
    g2_threshold_high = 0x1062,
    irq_latch = 0x1080,
    irq_level = 0x8004,
    irq_id = 0x8006,
    reset_module = 0x800A,
    serial_msb = 0x8180,
    serial_lsb = 0x8184
  };
  uint32_t frequency_to_pattern (float frequency) { return (std::max (0, std::min (0xFFFF, int(std::lround ((6.25e3 / frequency - 1) * 0xFFFF / (62.5e3 - 1)))))) << 16; }
}


ds50::V1495Driver::V1495Driver(ParameterSet const & ps): DS50FragmentGenerator(ps.get<ParameterSet> ("generator_ds50")) {
  mf::LogInfo("V1495") << "ds50daq version "
                       << ds50daq::GetPackageBuildInfo::getPackageBuildInfo().getPackageVersion()
                       << ", built "
                       << ds50daq::GetPackageBuildInfo::getPackageBuildInfo().getBuildTimestamp();

  statsHelper_.addMonitoredQuantityName(FRAGMENTS_PROCESSED_STAT_KEY);
  statsHelper_.addMonitoredQuantityName(READ_TIME_STAT_KEY);
  statsHelper_.addMonitoredQuantityName(FRAGMENT_CREATE_TIME_STAT_KEY);
  statsHelper_.addMonitoredQuantityName(REGISTER_RW_TIME_STAT_KEY);
  statsHelper_.addMonitoredQuantityName(IRQWAIT_TIME_STAT_KEY);
  statsHelper_.createCollectors(ps, 100, 30.0, 180.0);

  init_link (ps);

  init_metadata (ps);

  if (board_id () != ds50::Config::V1495_VETO_board_id) {
    init_trigger (ps);

    init_channels (ps);

    fr_type_ = Config::V1495_FRAGMENT_TYPE;
  } else {
    init_veto (ps);

    fr_type_ = Config::V1495VETO_FRAGMENT_TYPE;
  }

  dont_stop_ = ps.get<bool> ("dont_stop", false);
}

ds50::V1495Driver::~V1495Driver() noexcept {
  try {
    caen_io_->vme_sysreset ();
    sleep(1);
    close_link ();
  }
  catch (...) {
    mf::LogError ("V1495") << "Unknown exception caught in destructor.";
  }
}

void ds50::V1495Driver::close_link () {
  caen_io_ = 0;
}


void ds50::V1495Driver::init_link (ParameterSet const& ps) {
  caen_io_ = std::unique_ptr<detail::caen_io>(new detail::caen_io (cvA32_U_DATA, cvD32, ps.get<uint32_t> ("vme_base"), ps.get<std::string> ("link_type"), ps.get<int> ("link_number"), ps.get<int> ("link_node", 0)));

  caen_io_->write_register (V1495::reset_module, 1);
  usleep (100000);

  caen_io_->write_register (V1495::irq_level, 1);
  caen_io_->write_register (V1495::irq_id, 95);
  caen_io_->write_register (V1495::module_id, board_id ());

  irqwait_ms_ = ps.get<int> ("irqwait_ms", 200);
  irqwait_checks_ = ps.get<int> ("irqwait_checks", 2);
}

void ds50::V1495Driver::init_metadata (ParameterSet const& ps) {
  metadata_.serial = ((caen_io_->read_register (V1495::serial_msb, cvD16) & 0xff) << 8) | (caen_io_->read_register (V1495::serial_lsb, cvD16) & 0xff);
  metadata_.firmware_version = caen_io_->read_register (V1495::fw_revision);
  mf::LogInfo("V1495") << "board_id " << board_id () << " found trigger unit with serial " << std::hex << std::showbase <<  metadata_.serial << " fw release " << int(metadata_.firmware_version);

  metadata_.threshold = ps.get<bool> ("veto", false) ? 0 : ps.get<int> ("low_threshold");

  // 09-Oct-2014, KAB - added the use of the inhibit_us FHiCL parameter, but kept the
  // previous use of the gate width as the default (for backward compatibility)
  float inhibit_us = ps.get<float> ("inhibit_us",  -1.0F);
  if (inhibit_us < 0.0) {
    inhibit_us = std::max(ps.get<float> ("acquisition_gate_us"), 10.F);
    if (inhibit_us > 150) inhibit_us = (inhibit_us - 60) * 2 + 50;
  }
  metadata_.inhibit_20ns = inhibit_us / 20e-3;

  if (ps.get<bool> ("g2_mode", true)) {
    metadata_.g2_gate_20ns = std::min(ps.get<float> ("g2_gate_us", 5), 5.1F) / 20e-3;
    metadata_.g2_prescale = std::min(ps.get<int> ("g2_prescale", 33), 255);
    metadata_.g2_threshold = std::min(ps.get<int> ("g2_threshold", 360), 5080);
    metadata_.g2_threshold_high = std::min(ps.get<int> ("g2_threshold_high", 1500), 5080);
  } else metadata_.g2_threshold = metadata_.g2_prescale = metadata_.g2_gate_20ns = 0;
}

void ds50::V1495Driver::init_trigger (ParameterSet const& ps) {
  caen_io_->write_register (V1495::run_control, 1 << 1);
  caen_io_->write_register (V1495::window_inibit, metadata_.inhibit_20ns);

  float memory_full_extension_us = ps.get<float> ("memory_full_extension_us", 2);
  if (memory_full_extension_us > 0) {
    caen_io_->set_register_bits (V1495::run_control, 4);
    caen_io_->write_register (V1495::mf_extention, int(memory_full_extension_us / 20e-3));
  }

  caen_io_->clear_register_bits (V1495::trigger_control, (1 << 3) | 1);
  if (metadata_.threshold) {
    int low_threshold = std::min (abs(metadata_.threshold), 39);
    int high_threshold = std::min (ps.get<unsigned int> ("high_threshold", 39), 39U);
    int majority_time_10ns = std::min (ps.get<unsigned int> ("majority_time_10ns", 10), 15U);
    caen_io_->clear_register_bits (V1495::trigger_control, 0xffff0000);  // First reset the values
    if (metadata_.g2_threshold > 0) {
      caen_io_->write_register (V1495::g2_gate, metadata_.g2_gate_20ns);
      caen_io_->write_register (V1495::g2_threshold, metadata_.g2_threshold | (uint32_t(metadata_.g2_prescale) << 16));
      caen_io_->write_register (V1495::g2_threshold_high, metadata_.g2_threshold_high);
      caen_io_->set_register_bits (V1495::trigger_control, (1 << 3) | (majority_time_10ns << 16) | (low_threshold << 20) | (high_threshold << 26)); // G2 trigger
      std::cout << "G2 trigger programmed at gate " << int(metadata_.g2_gate_20ns) << " threshold " << int(metadata_.g2_threshold) << "-" << int(metadata_.g2_threshold_high) << " prescale " << int(metadata_.g2_prescale) << std::endl;
    } else caen_io_->set_register_bits (V1495::trigger_control, 1 | (majority_time_10ns << 16) | (low_threshold << 20) | (high_threshold << 26));
    if (metadata_.threshold < 0) caen_io_->write_register (V1495::run_control, 0);
  }

  caen_io_->set_register_bits (V1495::trigger_control, (1 << 9) | (1 << 8));  // enable external E4 and laser on E5

  bool random_triggers = ps.get<bool> ("random_triggers", false);
  if (random_triggers) caen_io_->set_register_bits (V1495::trigger_control, 1 << 2);
  else caen_io_->clear_register_bits (V1495::trigger_control, 1 << 2);

  float laser_frequency = ps.get<float> ("laser_frequency", -1);
  if (laser_frequency != 0) {
    if (laser_frequency > 0) caen_io_->write_register (V1495::laser_control, 1 | V1495::frequency_to_pattern (laser_frequency)); // No E5 input if internal mode is enabled
    else caen_io_->write_register (V1495::laser_control, 1 << 1);  // E5 enabled
    if (!metadata_.threshold) { 
      caen_io_->clear_register_bits (V1495::trigger_control, 0xffff0000);  // First reset the values
      caen_io_->set_register_bits (V1495::trigger_control, (10UL << 16) | (1UL << 20) | (39UL << 26)); // 100 us gate + 1 threshold + 39 High threshold
    } else if (ps.get<bool> ("g2_mode", true)) {
      mf::LogError ("V1495") << "g2_mode trigger and laser can not be enabled at the same time (disabling TPC triggers)";
      caen_io_->clear_register_bits (V1495::trigger_control, (1 << 3) | 1);
    }
  } else caen_io_->write_register (V1495::laser_control, 0); // laser completelly disabled

  float pulser_frequency = ps.get<float> ("pulser_frequency", -1);
  if (pulser_frequency > 0) {
    caen_io_->set_register_bits (V1495::run_control, V1495::frequency_to_pattern (pulser_frequency));
    caen_io_->set_register_bits (V1495::trigger_control, 1 << 1);
  } else caen_io_->clear_register_bits (V1495::trigger_control, 1 << 1);

  if (ps.get<bool> ("tpm_triggers", false)) {
    for (uint8_t i = 0; i < 8; i++) {
      uint32_t reg = ((uint16_t(1) << (i + 1)) - 1);
      caen_io_->write_register (V1495::test_pulser_dpm, reg << 16);
      caen_io_->write_register (V1495::test_pulser_dpm, (i * 0x100) << 16 | 1);
    }
    caen_io_->write_register (V1495::test_pulser_dpm, 0);
    caen_io_->set_register_bits (V1495::trigger_control, 1 << 10);
  } else caen_io_->clear_register_bits (V1495::trigger_control, 1 << 10);

  if (ps.get<bool> ("enable_veto_busy", false)) caen_io_->set_register_bits (V1495::run_control, 1 << 3);
  else caen_io_->clear_register_bits (V1495::run_control, 1 << 3);

  float tdc_pulse_delay_us = std::min(std::max(ps.get<float> ("tdc_pulse_delay_us", 0), 45e-3F), ps.get<float> ("acquisition_gate_us") - 1);
  caen_io_->write_register (V1495::tdc_delay, int(tdc_pulse_delay_us / 20e-3));

  if (ps.get<bool> ("fnal_mode", false)) caen_io_->set_register_bits (V1495::run_control, 1 << 15);
}


void ds50::V1495Driver::init_channels (ParameterSet const& ps) {
  std::vector<int> masked_channels = ps.get <std::vector<int>> ("masked_channels_0bias", std::vector<int>(0));
  uint64_t mask = 0;
  for (int ch: masked_channels) {
    if (ch <= 14) mask |= 1ULL << ch;
    else if (ch <= 30) mask |= 1ULL << (ch + 1);
    else mask |= 1ULL << (ch + 2);
  }

  if (mask) {
    caen_io_->write_register (V1495::port_a_mask, ~uint32_t(mask));
    caen_io_->write_register (V1495::port_b_mask, ~uint32_t(mask >> 32) & 0x3FF);
  }
}

void ds50::V1495Driver::init_veto (ParameterSet const& ps) {
  caen_io_->write_register (V1495::window_inibit, metadata_.inhibit_20ns);
  caen_io_->write_register (V1495::trigger_control, 0);
  caen_io_->write_register (V1495::run_control, 2);

  float tdc_pulse_delay_us = std::min(std::max(ps.get<float> ("tdc_pulse_delay_us", 0), 45e-3F), ps.get<float> ("acquisition_gate_us") - 1);
  caen_io_->write_register (V1495::tdc_delay_veto, int(tdc_pulse_delay_us / 20e-3));

    // No trigger from veto
  caen_io_->write_register (V1495::port_a_mask, 0);
  if (ps.get<bool> ("enable_tdc_busy", false))  caen_io_->write_register (V1495::port_b_mask, (3 << 16));
  else caen_io_->write_register (V1495::port_b_mask, 0);

  // Do trigger the veto?
  uint32_t fanout_vme_base = ps.get<uint32_t> ("fanout_vme_base") - ps.get<uint32_t> ("vme_base");
  // check if a veto_choice was specified
  int veto_choice = ps.get<int>("veto_choice", NO_CHOICE);
  if (veto_choice == NO_CHOICE) {
    // if not, default to VETO_OFF
    veto_choice = VETO_OFF;
    // check if the older-style of veto control was used
    if (ps.get<bool> ("enable_fanout_trigger_forward", false)) {
      veto_choice = ADC_SLAVE;
    }
  }
  if (veto_choice == VETO_OFF) {
    caen_io_->write_register (fanout_vme_base + V1495::run_control, 0);
  }
  else if (veto_choice == ADC_SLAVE) {
    caen_io_->write_register (fanout_vme_base + V1495::run_control, 0x81);
  }
  else if (veto_choice == HYBRID) {
    caen_io_->write_register (fanout_vme_base + V1495::run_control, 1 | (1UL << 7) | (3UL << 8) | (1 << 3));
  }
  else {
    mf::LogError ("V1495") << "board_id " << board_id () << ": unknown running mode (veto choice) = " << veto_choice;
  }
}

void ds50::V1495Driver::start_ () {
  caen_io_->write_register (V1495::run_number, run_number ());
  irqlost_ = 0;

  pause_ ();
  caen_io_->set_register_bits (V1495::run_control, 1);
  resume_ ();
  
  if (gettimeofday (&timing_, nullptr) < 0) mf::LogError ("V1495") << "board_id " << board_id () << " gettimeofday failed";
  timing_ev_ = 0;

  statsHelper_.resetStatistics();
}

void ds50::V1495Driver::stop_ () {
  if (!dont_stop_) {
    pause_ (); 
    caen_io_->clear_register_bits (V1495::run_control, 1);
    resume_ ();
  }
}

void ds50::V1495Driver::pause_ () {
  caen_io_->set_register_bits (V1495::run_control, 1 << 4);
  usleep (1000);
}

void ds50::V1495Driver::resume_ () {
  usleep (1000);
  caen_io_->clear_register_bits (V1495::run_control, 1 << 4);
}

bool ds50::V1495Driver::getNext__(artdaq::FragmentPtrs& frags) {
  bool will_stop = should_stop ();

  artdaq::MonitoredQuantity::TIME_POINT_T startTime;
  for (int j = 0; j < irqwait_checks_; j++) {
    uint32_t v;
    startTime = artdaq::MonitoredQuantity::getCurrentTime();
    bool no_irq = (!((will_stop && !(caen_io_->read_register (V1495::status) & 0x20)) || caen_io_->wait_irq (cvIRQ1, irqwait_ms_, v, cvD8)));
    statsHelper_.addSample(IRQWAIT_TIME_STAT_KEY,
                           (artdaq::MonitoredQuantity::getCurrentTime() - startTime));
    if (no_irq) continue;
    if ((v & 0xff) != 95) {
      mf::LogError ("V1495") << "board_id " << board_id () << " wrong IRQ vector " << std::showbase << std::hex << v;
      continue;
    }

    if (gettimeofday (&metadata_.trigger_time, nullptr) < 0) mf::LogError ("V1495") << "board_id " << board_id () << " gettimeofday failed";

    startTime = artdaq::MonitoredQuantity::getCurrentTime();
    std::unique_ptr<artdaq::Fragment> frag(new artdaq::Fragment(std::ceil (V1495Fragment::pod_size () / static_cast<double>(sizeof (artdaq::RawDataType))), ev_counter (), fragment_id (), fr_type_, metadata_));
    statsHelper_.addSample(FRAGMENT_CREATE_TIME_STAT_KEY,
                           (artdaq::MonitoredQuantity::getCurrentTime() - startTime));

    startTime = artdaq::MonitoredQuantity::getCurrentTime();
    uint32_t status_r = caen_io_->read_register (V1495::status);
    if ((status_r & 0x20)) mf::LogError ("V1495") << "board_id " << board_id () << " data not ready (status = " << std::showbase << std::hex << status_r << ")";
    statsHelper_.addSample(REGISTER_RW_TIME_STAT_KEY,
                           artdaq::MonitoredQuantity::getCurrentTime() - startTime);

    startTime = artdaq::MonitoredQuantity::getCurrentTime();
    try {
      int read_size = caen_io_->read_block (0x2000, &*frag->dataBegin (), V1495Fragment::pod_size (), detail::caen_io::sequential_fifo);
      if (read_size != int(V1495Fragment::pod_size ())) mf::LogWarning ("V1495") << "board_id " << board_id () << " Incomplete transfer " << read_size;
      statsHelper_.addSample(FRAGMENTS_PROCESSED_STAT_KEY, read_size);
    } catch (...) {
      mf::LogError ("V1495") << "board_id " << board_id () << " I/O error (status = " << std::showbase << std::hex << status_r << ") at event " << std::dec << ev_counter ();
      throw;
    }
    statsHelper_.addSample(READ_TIME_STAT_KEY,
                           artdaq::MonitoredQuantity::getCurrentTime() - startTime);

    startTime = artdaq::MonitoredQuantity::getCurrentTime();
    caen_io_->write_register (V1495::irq_latch, 1);
    statsHelper_.addSample(REGISTER_RW_TIME_STAT_KEY,
                           artdaq::MonitoredQuantity::getCurrentTime() - startTime);

    if (board_id () == ds50::Config::V1495_TPC_board_id) {
      float dt = difftime (metadata_.trigger_time.tv_sec, timing_.tv_sec);
      if (dt >= 30) {
	mf::LogInfo ("V1495") << "board_id " << board_id () << " processed " << ev_counter () - timing_ev_ << " events in " << dt << " at a rate of " << (ev_counter () - timing_ev_) / dt << " cps at event " << ev_counter ();
        timing_ = metadata_.trigger_time;
        timing_ev_ = ev_counter ();
      }
    }

#if DUMP_EVENTS
    std::cout << std::dec << board_id () << " Ev " << ev_counter () << "  " << std::hex << std::showbase << (((uint32_t*)&*frag->dataBegin ())[1] & 0xffff) << std::endl;
    std::cout << std::hex << std::showbase << "Run control " << caen_io_->read_register (V1495::run_control) << std::endl;
    std::cout << std::hex << std::showbase << "Trigger control " << caen_io_->read_register (V1495::trigger_control) << std::endl;
    std::cout << std::hex << std::showbase << "Status " << caen_io_->read_register (V1495::status) << std::endl;
    for (size_t i = 0; i < V1495Fragment::pod_size () / 4; i++) std::cout << std::hex << std::showbase << ((uint32_t*)&*frag->dataBegin ())[i] << std::endl;
    std::cout << "------------------------------" << std::endl;
#endif

    V1495Fragment const v1495_frag(*frag.get ());
    //if (!(ev_counter () % 4096))  std::cout << std::hex << std::showbase << v1495_frag << std::endl;

    if ((v1495_frag.live_time_100ns () == 0xFFFFFFFF) || (v1495_frag.incremental_inhibit_time_100ns () == 0xFFFFFFFF)) throw cet::exception("V1495") << "board_id " << board_id () << " live/dead time counters overun";

    if (board_id () != v1495_frag.board_id ()) mf::LogError ("V1495") << "board_id " << board_id () << " wrong module id found " << v1495_frag.board_id ();

    if ((v1495_frag.live_time_100ns () == 0xFFFFFFFF) || (v1495_frag.incremental_inhibit_time_100ns () == 0xFFFFFFFF)) throw cet::exception("V1495") << "board_id " << board_id () << " live/dead time counters overun";

    if ((v1495_frag.trigger_counter () & 0xfff) != v1495_frag.trigger_ordinal ()) mf::LogError ("V1495") << "board_id " << board_id () << " trigger ordinal " << std::hex << std::showbase << v1495_frag.trigger_ordinal () << " differs from trigger_counter " << v1495_frag.trigger_counter ();

    int ev_counter_diff = v1495_frag.trigger_counter () - ev_counter ();
    if (ev_counter_diff > 0) {
      mf::LogError ("V1495") << "board_id " << board_id () << " jump " << ev_counter_diff << " (got " << v1495_frag.trigger_counter () << ") at event " << ev_counter ();

      for (int i = 0; ev_counter_diff < 100 && i < ev_counter_diff; ++i) {
        startTime = artdaq::MonitoredQuantity::getCurrentTime();
        frags.emplace_back(new artdaq::Fragment(ev_counter () + i, fragment_id (), Config::MISSED_FRAGMENT_TYPE));
        statsHelper_.addSample(FRAGMENT_CREATE_TIME_STAT_KEY,
                               (artdaq::MonitoredQuantity::getCurrentTime() - startTime));
      }

      ev_counter_inc (ev_counter_diff);
      frag->setSequenceID (ev_counter ());
    } else if (ev_counter_diff < 0) {
      mf::LogError ("V1495") << "board_id " << board_id () << " duplicated fragment " << v1495_frag.trigger_counter () << " at event " << ev_counter ();
      continue;
    }

    frags.emplace_back (std::move (frag));
    ev_counter_inc ();
    if (statsHelper_.readyToReport(FRAGMENTS_PROCESSED_STAT_KEY, ev_counter())) {
      std::string statString = buildStatisticsString_();
      mf::LogDebug ("V1495") << statString;
    }
  }

  if (!frags.size () && will_stop) {
    // 16-Apr-2013, KAB:  added try/catch and single retry of the
    // status register read since it often fails the first time at
    // the Fermilab WH14 teststand.
    try {
      mf::LogDebug("V1495") << "board_id " << board_id () << " status is " << std::showbase << std::hex << caen_io_->read_register (V1495::status);
    }
    catch (...) {
      try {
        mf::LogDebug("V1495") << "board_id " << board_id () << " status is " << std::showbase << std::hex << caen_io_->read_register (V1495::status);
      }
      catch (...) {
        mf::LogDebug("V1495") << "board_id " << board_id () << " status is not readable.";
      }
    }
    return false;
  }
  return true;
}

std::string ds50::V1495Driver::buildStatisticsString_()
{
  std::ostringstream oss;
  oss << "V1495Driver statistics:" << std::endl << "  ";

  double fragmentCount = 1.0;
  double totalData = 1.0;
  artdaq::MonitoredQuantityPtr mqPtr = artdaq::StatisticsCollection::getInstance().
    getMonitoredQuantity(FRAGMENTS_PROCESSED_STAT_KEY);
  if (mqPtr.get() != 0) {
    artdaq::MonitoredQuantity::Stats stats;
    mqPtr->getStats(stats);
    fragmentCount = std::max(double(stats.recentSampleCount), 1.0);
    totalData = stats.recentValueSum;
  }

  double readRate = 0.0;
  double maxPossibleFragmentRate = 0.0;
  mqPtr = artdaq::StatisticsCollection::getInstance().
    getMonitoredQuantity(READ_TIME_STAT_KEY);
  if (mqPtr.get() != 0) {
    artdaq::MonitoredQuantity::Stats stats;
    mqPtr->getStats(stats);
    oss << "Average times per fragment: read time = "
        << (stats.recentValueSum / fragmentCount) << " sec";
    if (stats.recentValueSum > 0.0) {
      readRate = totalData / stats.recentValueSum;
    }
    if (stats.recentValueSum > 0.0) {
      maxPossibleFragmentRate = fragmentCount / stats.recentValueSum;
    }
  }

  mqPtr = artdaq::StatisticsCollection::getInstance().
    getMonitoredQuantity(IRQWAIT_TIME_STAT_KEY);
  if (mqPtr.get() != 0) {
    artdaq::MonitoredQuantity::Stats stats;
    mqPtr->getStats(stats);
    oss << ", IRQ wait time = "
        << (stats.recentValueSum / fragmentCount) << " sec";
  }

  mqPtr = artdaq::StatisticsCollection::getInstance().
    getMonitoredQuantity(FRAGMENT_CREATE_TIME_STAT_KEY);
  if (mqPtr.get() != 0) {
    artdaq::MonitoredQuantity::Stats stats;
    mqPtr->getStats(stats);
    oss << ", Fragment creation time = "
        << (stats.recentValueSum / fragmentCount) << " sec";
  }

  mqPtr = artdaq::StatisticsCollection::getInstance().
    getMonitoredQuantity(REGISTER_RW_TIME_STAT_KEY);
  if (mqPtr.get() != 0) {
    artdaq::MonitoredQuantity::Stats stats;
    mqPtr->getStats(stats);
    oss << ", register read/write time = "
        << (stats.recentValueSum / fragmentCount) << " sec";
  }

  oss << std::endl;
  oss << "  Average read rate = " << (readRate / 1024.0 / 1024.0) << " MB/sec";
  oss << ", maximum possible fragment rate = " << maxPossibleFragmentRate << " Hz";

  return oss.str();
}

DEFINE_ARTDAQ_COMMANDABLE_GENERATOR(ds50::V1495Driver)
